"""
:mod:`edx_ace.channel.braze` implements a Braze-based email delivery channel for ACE.
"""
import logging
import random
import warnings
from datetime import timedelta
from gettext import gettext as _

import requests

from django.conf import settings

from edx_ace.channel import Channel
from edx_ace.channel.django_email import DjangoEmailChannel
from edx_ace.channel.mixins import EmailChannelMixin
from edx_ace.errors import FatalChannelDeliveryError, RecoverableChannelDeliveryError
from edx_ace.utils.date import get_current_time

LOG = logging.getLogger(__name__)

NEXT_ATTEMPT_DELAY_SECONDS = 30
BRAZE_API_TIMEOUT = 5


class BrazeEmailChannel(EmailChannelMixin, Channel):
    """
    An email channel for delivering messages to users using Braze.

    This channel makes use of the Braze REST API to send messages. It is designed for "at most once" delivery of
    messages. It will make a reasonable attempt to deliver the message and give up if it can't. It also only confirms
    that Braze has received the request to send the email, it doesn't actually confirm that it made it to the
    recipient.

    See the Braze documentation for message sending, for more information:
    https://www.braze.com/docs/api/endpoints/messaging/send_messages/post_send_messages/

    The recipient email address is ignored, instead Braze uses its stored email address for the recipient. Although,
    if the lms_user_id is not valid, this channel falls back to the Django email channel and then the address will
    be used.

    The integration with Braze requires several Django settings to be defined.

    The ACE_CHANNEL_BRAZE_CAMPAIGNS setting is optional, but if it is defined, it should be a mapping of ACE message
    names to campaign ids. And optionally a message variation id, separated by a colon. See the example below.

    Example:

        Sample settings::

            .. settings_start
            ACE_CHANNEL_BRAZE_API_KEY = "1c304d0d-c800-4da3-bfaa-41b1189b34cb"
            ACE_CHANNEL_BRAZE_APP_ID = "f6232495-6ad6-4310-bab5-5673768856aa"
            ACE_CHANNEL_BRAZE_FROM_EMAIL = "Open edX <openedx@example.com>"
            ACE_CHANNEL_BRAZE_REST_ENDPOINT = "rest.iad-01.braze.com"
            ACE_CHANNEL_BRAZE_CAMPAIGNS = {
                "deletionnotificationmessage": "campaign_id:variation_id"
            }
            .. settings_end
    """

    _API_KEY_SETTING = 'ACE_CHANNEL_BRAZE_API_KEY'
    _APP_ID_SETTING = 'ACE_CHANNEL_BRAZE_APP_ID'
    _CAMPAIGNS_SETTING = 'ACE_CHANNEL_BRAZE_CAMPAIGNS'  # optional
    _ENDPOINT_SETTING = 'ACE_CHANNEL_BRAZE_REST_ENDPOINT'
    _FROM_EMAIL_SETTING = 'ACE_CHANNEL_BRAZE_FROM_EMAIL'  # optional

    @classmethod
    def enabled(cls):
        """
        Returns: True iff all required settings are not empty and the Braze client library is installed.
        """
        ok = True

        for setting in (
            cls._API_KEY_SETTING,
            cls._APP_ID_SETTING,
            cls._ENDPOINT_SETTING,
        ):
            if not getattr(settings, setting, None):
                ok = False
                LOG.warning('%s is not set, Braze email channel is disabled.', setting)

        return ok

    @property
    def action_links(self):
        """
        This method is now deprecated in favor of get_action_links,
        but will continue to work for the time being as it calls get_action_links under the hood.
        """
        warnings.warn("This method is now deprecated in favor of get_action_links")
        return self.get_action_links()

    def get_action_links(self, **kwargs):
        """
        Provides list of action links, called by templates directly.
        Supported kwargs:
            omit_unsubscribe_link (bool): Removes the unsubscribe link from the email.
                DO NOT send emails with no unsubscribe link unless you are sure it will not violate the CANSPAM act.
        """
        omit_unsubscribe_link = kwargs.get('omit_unsubscribe_link')
        action_links = []
        if not omit_unsubscribe_link:
            action_links += [('{{${set_user_to_unsubscribed_url}}}', _('Unsubscribe from this list'))]
        return action_links

    @property
    def tracker_image_sources(self):
        """Provides list of trackers, called by templates directly"""
        return []  # not needed

    def deliver(self, message, rendered_message):
        if not self.enabled():
            raise FatalChannelDeliveryError('Braze channel is disabled, unable to send')

        if not message.recipient.lms_user_id:
            # This channel assumes that you have Braze configured with LMS user_ids as your external_user_id in Braze.
            # Unfortunately, that means that we can't send emails to users that aren't registered with the LMS,
            # which some callers of ACE may attempt to do (despite the lms_user_id being a required Recipient field).
            # In these cases, we fall back to a simple Django smtp email.
            DjangoEmailChannel().deliver(message, rendered_message)
            return

        transactional = message.options.get('transactional', False)
        override_frequency_capping = message.options.get('override_frequency_capping', transactional)
        body_html = self.make_simple_html_template(rendered_message.head_html, rendered_message.body_html)

        # Allow our settings to override the from address, because Braze requires specific configured from addresses,
        # which are tied to specific ip addresses that are "ip warmed" to help delivery of the emails not get sent
        # to promotional/spam inboxes.
        from_address = getattr(settings, self._FROM_EMAIL_SETTING, None) or self.get_from_address(message)

        logger = message.get_message_specific_logger(LOG)
        logger.debug('Sending to Braze')

        # https://www.braze.com/docs/api/endpoints/messaging/send_messages/post_send_messages/
        # https://www.braze.com/docs/api/objects_filters/email_object/
        response = requests.post(
            self._send_url(),
            headers=self._auth_headers(),
            json={
                'external_user_ids': [str(message.recipient.lms_user_id)],
                'recipient_subscription_state': 'all' if transactional else 'subscribed',
                'campaign_id': self._campaign_id(message.name),
                'override_frequency_capping': override_frequency_capping,
                'messages': {
                    'email': {
                        'app_id': getattr(settings, self._APP_ID_SETTING),
                        'subject': self.get_subject(rendered_message),
                        'from': from_address,
                        'reply_to': message.options.get('reply_to'),
                        'body': body_html,
                        'plaintext_body': rendered_message.body,
                        'message_variation_id': self._variation_id(message.name),
                        'should_inline_css': False,  # this feature messes with inline CSS already in ACE templates
                    },
                },
            },
            timeout=getattr(settings, 'ACE_DEFAULT_API_TIMEOUT', BRAZE_API_TIMEOUT)
        )

        try:
            response.raise_for_status()
            logger.debug('Successfully sent to Braze (dispatch ID %s)', response.json()['dispatch_id'])

        except requests.exceptions.HTTPError as exc:
            # https://www.braze.com/docs/api/errors/
            message = response.json().get('message', 'Unknown error')
            logger.debug('Failed to send to Braze: %s', message)
            self._handle_error_response(response, message, exc)

    def overrides_delivery_for_message(self, message):
        # If we have a campaign configured for this message, let's deliver it ourselves, even if it's a transactional
        # message. Presumably that campaign is set up to ignore global delivery caps so that we don't drop such a
        # transactional message on the floor. This kind of configuration could be done to get nice email metrics.
        return self._campaign_id(message.name) is not None

    def _handle_error_response(self, response, message, exception):
        """
        Handle an error response from Braze, either by retrying or failing
        with an appropriate exception.

        Arguments:
            response: The HTTP response received from Braze.
            message: An error message from Braze.
            exception: The exception that triggered this error.
        """
        if response.status_code == 429 or 500 <= response.status_code < 600:
            next_attempt_time = get_current_time() + timedelta(
                seconds=NEXT_ATTEMPT_DELAY_SECONDS + random.uniform(-2, 2)
            )
            raise RecoverableChannelDeliveryError(
                f'Recoverable Braze error (status_code={response.status_code}): {message}',
                next_attempt_time
            ) from exception

        raise FatalChannelDeliveryError(
            f'Fatal Braze error (status_code={response.status_code}): {message}'
        ) from exception

    @classmethod
    def _auth_headers(cls):
        """Returns authorization headers suitable for passing to the requests library"""
        return {
            'Authorization': 'Bearer ' + getattr(settings, cls._API_KEY_SETTING),
        }

    @classmethod
    def _send_url(cls):
        """Returns the send-message API URL"""
        endpoint = getattr(settings, cls._ENDPOINT_SETTING)
        return f'https://{endpoint}/messages/send'

    @classmethod
    def _campaign_id(cls, name):
        """Returns the campaign ID for a given ACE message name or None if no match is found"""
        campaign = getattr(settings, cls._CAMPAIGNS_SETTING, {}).get(name)
        if campaign:
            return campaign.split(':')[0]
        return None

    @classmethod
    def _variation_id(cls, name):
        """Returns the variation ID for a given ACE message name or None if no match is found"""
        campaign = getattr(settings, cls._CAMPAIGNS_SETTING, {}).get(name)
        if campaign:
            campaign_parts = campaign.split(':')
            if len(campaign_parts) > 1:
                return campaign_parts[1]
        return None



"""
:mod:`edx_ace.errors` exposes all exceptions that are specific to ACE.
"""


# TODO(later): To prevent a monolith of exception classes here,
#   we may want to declare Channel-related errors in the
#   channel subdirectory.


class ChannelError(Exception):
    """Indicates something went wrong in a delivery channel."""
    pass


class RecoverableChannelDeliveryError(ChannelError):
    """An error occurred during channel delivery that is non-fatal. The caller should re-attempt at a later time."""

    def __init__(self, message, next_attempt_time):
        self.next_attempt_time = next_attempt_time
        super().__init__(message)


class FatalChannelDeliveryError(ChannelError):
    """A fatal error occurred during channel delivery. Do not retry."""
    pass


class UnsupportedChannelError(ChannelError):
    """Raised when an attempt is made to process a message for an unsupported channel."""
    pass


class InvalidMessageError(Exception):
    """Encountered a message that cannot be sent due to missing or inconsistent information."""
    pass



"""
:mod:`edx_ace.monitoring` exposes functions that are useful for reporting ACE
message delivery stats to monitoring services.
"""
from edx_django_utils.monitoring import set_custom_attribute


def report(key, value):
    set_custom_attribute(key, value)



"""
:mod:`edx_ace.utils.plugins` contains utility functions used
to make working with the ACE plugin system easier. These are intended
for internal use by ACE.
"""
import logging
from functools import partial

from stevedore import enabled

LOG = logging.getLogger(__name__)


def get_manager(namespace, names=None):
    """
    Get the stevedore extension manager for this namespace.

    Args:
        namespace (basestring): The entry point namespace to load plugins for.
        names (list): A list of names to load. If this is ``None`` then all extension will be loaded from this
            namespace.

    Returns:
        stevedore.enabled.EnabledExtensionManager: Extension manager with all extensions instantiated.
    """
    return enabled.EnabledExtensionManager(
        namespace=namespace,
        check_func=partial(check_plugin, namespace=namespace, names=names),
        invoke_on_load=True,
    )


def get_plugins(namespace, names=None):
    """
    Get all extensions for this namespace and list of names.

    Args:
        namespace (basestring): The entry point namespace to load plugins for.
        names (list): A list of names to load. If this is ``None`` then all extension will be loaded from this
            namespace.

    Returns:
        list: A list of extensions.
    """
    return list(get_manager(namespace, names))


def check_plugin(extension, namespace, names=None):
    """
    Check the extension to see if it's enabled.

    Args:
        extension (stevedore.extension.Extension): The extension to check.
        namespace (basestring): The namespace that the extension was loaded from.
        names (list): A whitelist of extensions that should be checked.

    Returns:
        bool: Whether or not this extension is enabled and should be used.
    """
    if names is None or extension.name in names:
        plugin_enabled = extension.plugin.enabled()
        if not plugin_enabled:
            LOG.info('Extension with name %s for namespace %s is not enabled', extension.name, namespace)
        return plugin_enabled
    return False



"""
ACE (Automated Communications Engine) is a framework for automatically
sending messages to users.

:mod:`edx_ace` exports the typical set of functions and classes needed to use
ACE.
"""

from .ace import send
from .channel import Channel, ChannelType
from .message import Message, MessageType
from .policy import Policy, PolicyResult
from .recipient import Recipient
from .recipient_resolver import RecipientResolver

__version__ = '1.11.3'


__all__ = [
    'send',
    '__version__',
    'Message',
    'MessageType',
    'Recipient',
    'RecipientResolver',
    'ChannelType',
    'Channel',
    'Policy',
    'PolicyResult',
]



"""
Tests of :mod:`edx_ace.channel`.
"""
from unittest.mock import patch

from django.test import TestCase, override_settings

from edx_ace.channel import ChannelMap, ChannelType, get_channel_for_message
from edx_ace.channel.braze import BrazeEmailChannel
from edx_ace.channel.file import FileEmailChannel
from edx_ace.channel.push_notification import PushNotificationChannel
from edx_ace.errors import UnsupportedChannelError
from edx_ace.message import Message
from edx_ace.recipient import Recipient
from edx_ace.utils.date import get_current_time


class TestChannelMap(TestCase):
    """
    Tests for the channels().
    """

    def setUp(self):
        self.msg_kwargs = {
            'app_label': 'test_app_label',
            'name': 'test_message',
            'expiration_time': get_current_time(),
            'context': {
                'key1': 'value1',
                'key2': 'value2',
            },
            'recipient': Recipient(
                lms_user_id=123,
            )
        }

    @override_settings(
        ACE_CHANNEL_BRAZE_CAMPAIGNS={
            'campaign_msg': 'campaign_id:variation_id',
        },
        ACE_CHANNEL_DEFAULT_EMAIL='braze_email',
        ACE_CHANNEL_TRANSACTIONAL_EMAIL='file_email',
        ACE_CHANNEL_DEFAULT_PUSH='push_notification',
    )
    def test_get_channel_for_message(self):
        channel_map = ChannelMap([
            ['file_email', FileEmailChannel()],
            ['braze_email', BrazeEmailChannel()],
            ['push_notifications', PushNotificationChannel()],
        ])

        transactional_msg = Message(options={'transactional': True}, **self.msg_kwargs)
        info_msg = Message(options={}, **self.msg_kwargs)

        # A transactional message which the default channel wants to handle and thus should be directed appropriately
        transactional_campaign_msg = Message(options={'transactional': True}, **self.msg_kwargs)
        transactional_campaign_msg.name = 'campaign_msg'

        with patch('edx_ace.channel.channels', return_value=channel_map):
            assert isinstance(get_channel_for_message(ChannelType.EMAIL, transactional_msg), FileEmailChannel)
            assert isinstance(get_channel_for_message(ChannelType.EMAIL, transactional_campaign_msg), BrazeEmailChannel)
            assert isinstance(get_channel_for_message(ChannelType.EMAIL, info_msg), BrazeEmailChannel)
            assert isinstance(get_channel_for_message(ChannelType.PUSH, info_msg), PushNotificationChannel)

            with self.assertRaises(UnsupportedChannelError):
                assert get_channel_for_message('unsupported_channel_type', transactional_msg)

    @override_settings(
        ACE_CHANNEL_DEFAULT_EMAIL='braze_email',
        ACE_CHANNEL_TRANSACTIONAL_EMAIL='file_email',
    )
    def test_default_channel(self):
        channel_map = ChannelMap([
            ['braze', BrazeEmailChannel()],
        ])

        message = Message(options={'transactional': True}, **self.msg_kwargs)

        with patch('edx_ace.channel.channels', return_value=channel_map):
            channel = get_channel_for_message(ChannelType.EMAIL, message)
            assert isinstance(channel, BrazeEmailChannel)



"""
Internal module configuring ACE as a Django app.
"""
from django.apps import AppConfig


class EdxAceConfig(AppConfig):
    """
    Configuration for the edx_ace Django application.
    """

    name = 'edx_ace'



"""
Tests of :mod:`edx_ace.ace`.
"""
from unittest.mock import Mock, patch

from django.template import TemplateDoesNotExist
from django.test import TestCase

from edx_ace import ace
from edx_ace.channel import ChannelMap, ChannelType
from edx_ace.errors import UnsupportedChannelError
from edx_ace.message import Message
from edx_ace.recipient import Recipient
from edx_ace.renderers import RenderedEmail
from edx_ace.test_utils import StubPolicy, patch_policies


class TestAce(TestCase):
    """
    Tests for the send method.
    """
    def test_ace_send_happy_path(self):
        patch_policies(self, [StubPolicy([ChannelType.PUSH])])
        mock_channel = Mock(
            channel_type=ChannelType.EMAIL,
            action_links=[],
            get_action_links=[],
            tracker_image_sources=[],
        )

        recipient = Recipient(lms_user_id=123)
        msg = Message(
            app_label='testapp',
            name='testmessage',
            recipient=recipient,
        )

        channel_map = ChannelMap([
            ['sailthru_email', mock_channel],
        ])

        with patch('edx_ace.channel.channels', return_value=channel_map):
            ace.send(msg)

        mock_channel.deliver.assert_called_once_with(
            msg,
            RenderedEmail(
                from_name='template from_name.txt',
                subject='template subject.txt',
                # The new lines are needed because the template has some tags, which means leftover newlines
                body_html='template body.html\n\n\n\n\n',
                head_html='template head.html\n',
                body='template body.txt',
            ),
        )

    @patch('edx_ace.ace.get_channel_for_message', side_effect=UnsupportedChannelError)
    def test_ace_send_unsupported_channel(self, *_args):
        recipient = Recipient(lms_user_id=123)
        msg = Message(
            app_label='testapp',
            name='testmessage',
            recipient=recipient,
        )

        ace.send(msg)  # UnsupportedChannelError shouldn't throw UnsupportedChannelError

    @patch('edx_ace.ace.log')
    @patch('edx_ace.ace.policy.channels_for', return_value=[ChannelType.PUSH])
    def test_ace_send_skip_limited_channel(self, mock_channels_for, mock_log):
        recipient = Recipient(lms_user_id=123)
        msg = Message(
            app_label='testapp',
            name='testmessage',
            recipient=recipient,
        )

        ace.send(msg, limit_to_channels=[ChannelType.EMAIL])

        mock_channels_for.assert_called_once_with(msg)
        mock_log.debug.assert_called_once_with('Skipping channel %s', ChannelType.PUSH)

    @patch('edx_ace.ace.presentation.render', side_effect=TemplateDoesNotExist('template not found'))
    def test_ace_send_template_does_not_exists(self, *_args):
        recipient = Recipient(lms_user_id=123)
        report_mock = Mock()

        msg = Mock(
            app_label='testapp',
            name='testmessage',
            recipient=recipient,
            context={},
            report=report_mock,
        )
        ace.send(msg)
        report_mock.assert_called_with(
            'template_error',
            'Unable to send message because template not found\ntemplate not found'
        )



"""
Channel for sending push notifications.
"""
import logging
import re

from firebase_admin.messaging import APNSConfig, APNSPayload, Aps, ApsAlert
from push_notifications.gcm import dict_to_fcm_message, send_message
from push_notifications.models import GCMDevice

from django.conf import settings

from edx_ace.channel import Channel, ChannelType
from edx_ace.errors import FatalChannelDeliveryError
from edx_ace.message import Message
from edx_ace.renderers import RenderedPushNotification

LOG = logging.getLogger(__name__)
APNS_DEFAULT_PRIORITY = '5'
APNS_DEFAULT_PUSH_TYPE = 'alert'


class PushNotificationChannel(Channel):
    """
    A channel for sending push notifications.
    """

    channel_type = ChannelType.PUSH

    @classmethod
    def enabled(cls):
        """
        Returns true if the push notification settings are configured.
        """
        return bool(getattr(settings, 'PUSH_NOTIFICATIONS_SETTINGS', None))

    def deliver(self, message: Message, rendered_message: RenderedPushNotification) -> None:
        """
        Transmit a rendered message to a recipient.

        Args:
            message: The message to transmit.
            rendered_message: The rendered content of the message that has been personalized
                for this particular recipient.
        """
        device_tokens = self.get_user_device_tokens(message.recipient.lms_user_id)
        if not device_tokens:
            LOG.info(
                'Recipient with ID %s has no push token. Skipping push notification.',
                message.recipient.lms_user_id
            )
            return

        for token in device_tokens:
            self.send_message(message, token, rendered_message)

    def send_message(self, message: Message, token: str, rendered_message: RenderedPushNotification) -> None:
        """
        Send a push notification to a device by token.
        """
        notification_data = {
            'title': self.compress_spaces(rendered_message.title),
            'body': self.compress_spaces(rendered_message.body),
            'notification_key': token,
            **message.context.get('push_notification_extra_context', {}),
        }
        message = dict_to_fcm_message(notification_data)
        # Note: By default dict_to_fcm_message does not support APNS configuration,
        #   only Android configuration, so we need to collect and set it manually.
        apns_config = self.collect_apns_config(notification_data)
        message.apns = apns_config
        try:
            send_message(token, message, settings.FCM_APP_NAME)
        except Exception as e:
            LOG.exception('Failed to send push notification to %s', token)
            raise FatalChannelDeliveryError(f'Failed to send push notification to {token}') from e

    @staticmethod
    def collect_apns_config(notification_data: dict) -> APNSConfig:
        """
        Collect APNS configuration with payload for the push notification.

        This APNSConfig must be set to notifications for Firebase to send push notifications to iOS devices.
        Notification has default priority and visibility settings, described in Apple's documentation.
        (https://developer.apple.com/documentation/usernotifications/sending-notification-requests-to-apns)
        """
        apns_alert = ApsAlert(title=notification_data['title'], body=notification_data['body'])
        aps = Aps(alert=apns_alert, sound='default')
        return APNSConfig(
            headers={'apns-priority': APNS_DEFAULT_PRIORITY, 'apns-push-type': APNS_DEFAULT_PUSH_TYPE},
            payload=APNSPayload(aps)
        )

    @staticmethod
    def get_user_device_tokens(user_id: int) -> list:
        """
        Get the device tokens for a user.
        """
        return list(GCMDevice.objects.filter(
            user_id=user_id,
            cloud_message_type='FCM',
            active=True,
        ).values_list('registration_id', flat=True))

    @staticmethod
    def compress_spaces(html_str: str) -> str:
        """
        Compress spaces and remove newlines to make it easier to author templates.
        """
        return re.sub('\\s+', ' ', html_str, re.UNICODE).strip()



"""
Utils for signals.
"""
from edx_ace.signals import ACE_MESSAGE_SENT


def make_serializable_object(obj):
    """
    Takes a dictionary/list and returns a dictionary/list with all the values converted
    to JSON serializable objects.
    """
    try:
        if isinstance(obj, (int, float, str, bool)) or obj is None:
            return obj
        elif isinstance(obj, dict):
            return {key: make_serializable_object(value) for key, value in obj.items()}
        elif isinstance(obj, list):
            return [make_serializable_object(element) for element in obj]
    except Exception:    # pylint: disable=broad-except
        pass
    return str(obj)


def send_ace_message_sent_signal(channel, message):
    """
    Creates dictionary from message, makes it JSON serializable and
    sends the ACE_MESSAGE_SENT signal.
    """
    try:
        channel_name = channel.__class__.__name__
    except AttributeError:
        channel_name = 'Other'
    data = {
        'name': message.name,
        'app_label': message.app_label,
        'recipient': {
            'email': getattr(message.recipient, 'email_address', ''),
            'user_id': getattr(message.recipient, 'lms_user_id', ''),
        },
        'channel': channel_name,
        'context': message.context,
        'options': message.options,
        'uuid': str(message.uuid),
        'send_uuid': str(message.send_uuid),
    }
    ACE_MESSAGE_SENT.send(sender=channel, message=make_serializable_object(data))



# pylint: disable=missing-docstring
from unittest.mock import patch

import ddt

from django.test import TestCase, override_settings

from edx_ace.channel.sailthru import SailthruEmailChannel
from edx_ace.delivery import deliver
from edx_ace.errors import InvalidMessageError
from edx_ace.message import Message
from edx_ace.presentation import render
from edx_ace.recipient import Recipient


@ddt.ddt
class TestSailthruChannel(TestCase):

    def setUp(self):
        self.channel = SailthruEmailChannel()

    def test_render_email_with_sailthru(self):
        message = Message(
            app_label='testapp',
            name='testmessage',
            options={},
            recipient=Recipient(lms_user_id=123, email_address='mr@robot.io'),
        )

        rendered_email = render(self.channel, message)

        assert '{beacon_src}' in rendered_email.body_html
        assert '{view_url}' in rendered_email.body_html
        assert '{optout_confirm_url}' in rendered_email.body_html

    @override_settings(
        ACE_CHANNEL_SAILTHRU_DEBUG=False,
    )
    @ddt.data(
        ({'from_address': 'custom@example.com'}, {'behalf_email': 'custom@example.com'}),
        ({'reply_to': 'student@example.com'}, {'behalf_email': 'student@example.com'}),
        ({'irrelevant': 'test'}, {}),
        ({}, {}),
    )
    @ddt.unpack
    def test_on_behalf_option_with_sailthru(self, message_options, expected_options):
        """
        Tests sailthru send API is called with on_behalf option
        """
        message = Message(
            app_label='testapp',
            name='testmessage',
            options=message_options,
            recipient=Recipient(lms_user_id=123, email_address='mr@robot.io'),
        )
        rendered_email = render(self.channel, message)

        with patch('edx_ace.channel.sailthru.SailthruClient.send') as mock_send:
            deliver(self.channel, rendered_email, message)
            self.assertEqual(mock_send.call_args_list[0][1]['options'], expected_options)

    def test_no_recipient_email_address(self):
        message = Message(
            app_label='testapp',
            name='testmessage',
            options={},
            recipient=Recipient(lms_user_id=123),
        )
        rendered_email = render(self.channel, message)

        with self.assertRaisesRegex(InvalidMessageError, 'No email address'):
            deliver(self.channel, rendered_email, message)

    def test_get_action_links_omit_unsubscribe_link(self):
        """Some emails use a setting that omits the unsubscribe action link"""
        assert len(self.channel.action_links) == 2
        assert len(self.channel.get_action_links(omit_unsubscribe_link=True)) == 1

    def test_email_omits_unsubscribe_link(self):
        """Basic email send, no special settings"""
        message = Message(
            app_label='testapp',
            name='testmessage',
            options={},
            context={'omit_unsubscribe_link': True},
            recipient=Recipient(lms_user_id=123, email_address='mr@robot.io'),
        )

        rendered_email = render(self.channel, message)

        assert '{optout_confirm_url}' not in rendered_email.body_html



"""
Tests for the utils/signals module.
"""
from django.test import TestCase

from edx_ace.utils.signals import make_serializable_object


class TestMakeSerializableObject(TestCase):
    def test_primitive_types(self):
        self.assertEqual(make_serializable_object(42), 42)
        self.assertEqual(make_serializable_object(3.14), 3.14)
        self.assertEqual(make_serializable_object("string"), "string")
        self.assertEqual(make_serializable_object(True), True)
        self.assertEqual(make_serializable_object(None), None)

    def test_dict(self):
        input_dict = {
            "int": 1,
            "float": 2.0,
            "str": "test",
            "bool": False,
            "none": None,
            "list": [1, 2, 3],
            "nested_dict": {"key": "value"}
        }
        self.assertEqual(make_serializable_object(input_dict), input_dict)

    def test_list(self):
        input_list = [1, 2.0, "test", False, None, [1, 2, 3], {"key": "value"}]
        self.assertEqual(make_serializable_object(input_list), input_list)

    def test_non_serializable(self):
        class NonSerializable:
            pass

        obj = NonSerializable()
        self.assertEqual(make_serializable_object(obj), str(obj))

    def test_non_serializable_list(self):
        class NonSerializable:
            pass

        obj = NonSerializable()
        obj2 = NonSerializable()
        obj_list = [obj, obj2]
        self.assertEqual(make_serializable_object(obj_list), [str(obj), str(obj2)])



"""
Functions for delivering ACE messages.

This is an internal interface used by :func:`.ace.send`.
"""
import datetime
import logging
import time

from django.conf import settings

from edx_ace.errors import RecoverableChannelDeliveryError
from edx_ace.utils.date import get_current_time
from edx_ace.utils.signals import send_ace_message_sent_signal

LOG = logging.getLogger(__name__)

# Set a maximum expiration delay, for now, since we're currently
# not using celery to re-enqueue retries and so we could be
# potentially blocking a worker indefinitely.
#
# TODO(later): Use celery per channel delivery to be smarter
# about re-enqueueing and retrying, and to not block workers.
MAX_EXPIRATION_DELAY = 5 * 60


def deliver(channel, rendered_message, message):
    """
    Deliver a message via a particular channel.

    Args:
        channel (Channel): The channel to deliver the message over.
        rendered_message (object): Each attribute of this object contains rendered content.
        message (Message): The message that is being sent.

    Raises:
        :class:`.UnsupportedChannelError`: If no channel of the requested channel type is available.

    """
    logger = message.get_message_specific_logger(LOG)
    channel_type = channel.channel_type

    timeout_seconds = getattr(settings, 'ACE_DEFAULT_EXPIRATION_DELAY', 120)
    start_time = get_current_time()
    default_expiration_time = start_time + datetime.timedelta(seconds=timeout_seconds)
    max_expiration_time = start_time + datetime.timedelta(seconds=MAX_EXPIRATION_DELAY)
    expiration_time = min(max_expiration_time, message.expiration_time or default_expiration_time)

    logger.debug('Attempting delivery of message')
    while get_current_time() < expiration_time:
        try:
            channel.deliver(message, rendered_message)
        except RecoverableChannelDeliveryError as delivery_error:
            num_seconds = (delivery_error.next_attempt_time - get_current_time()).total_seconds()
            logger.debug('Encountered a recoverable delivery error.')
            if delivery_error.next_attempt_time > expiration_time:
                logger.debug('Message will expire before delivery can be reattempted, aborting.')
                break
            logger.debug('Sleeping for %d seconds before reattempting delivery of message.', num_seconds)
            time.sleep(num_seconds)
            message.report(f'{channel_type}_delivery_retried', num_seconds)
        else:
            message.report(f'{channel_type}_delivery_succeeded', True)
            send_ace_message_sent_signal(channel, message)
            return

    delivery_expired_report = f'{channel_type}_delivery_expired'
    logger.debug(delivery_expired_report)
    message.report(delivery_expired_report, get_current_time() - start_time)



"""
:mod:`edx_ace.serialization` contains :class:`MessageAttributeSerializationMixin`,
which allows messages to be round-tripped through JSON, and
:class:`MessageEncoder`, which actually performs the JSON encoding.
"""
import json
from uuid import UUID

import attr

from django.utils.functional import Promise

from edx_ace.utils import date


class MessageAttributeSerializationMixin:
    """
    This mixin allows an object to be serialized to (and deserialized from)
    a JSON string.

    :meth:`__str__` and :meth:`from_string` function as inverses,
    and are the primary point of interaction with this mixin by
    outside clients.

    :meth:`to_json` is used to recursively convert the object to a
    python dictionary that can then be encoded to a JSON string.
    """
    def __str__(self):
        return json.dumps(self, cls=MessageEncoder)

    @classmethod
    def from_string(cls, string_value):
        """
        Decode a JSON-encoded string representation of this type.

        Arguments:
            string_value (str): The JSON string to decode.

        Returns:
            An instance of this class.
        """
        fields = json.loads(
            string_value,
            object_hook=cls._deserialize,
        )
        uuid = fields.pop('uuid')
        instance = cls(**fields)
        instance.uuid = uuid
        return instance

    def to_json(self):
        """
        Returns: dict
            a python dictionary containing all serializable fields
            of this object, suitable for JSON-encoding.
        """
        return attr.asdict(self)

    @classmethod
    def _deserialize(cls, json_value):
        """
        Deserialize a JSON value (a python dictionary) into an
        instance of the current class.

        Arguments:
            json_value (dict): The json object to decode.

        Returns:
            An instance of the current class.
        """
        fields = json_value.copy()
        for field_name, field_value in fields.items():
            fields[field_name] = cls._deserialize_field(field_name, field_value)
        return fields

    @classmethod
    def _deserialize_field(cls, field_name, field_value):
        """
        Deserialize a single encoded field value for this message.

        Arguments:
            field_name (str): The name of the field being decoded.
            field_value: The value of field being deserialized.
        """
        # We have to import these here to avoid a circular dependency since these classes use this mixin.
        from edx_ace.message import Message  # pylint: disable=import-outside-toplevel
        from edx_ace.recipient import Recipient  # pylint: disable=import-outside-toplevel

        if field_value is None:
            return None

        if field_name == 'expiration_time':
            return date.deserialize(field_value)
        elif field_name in ('uuid', 'send_uuid'):
            return UUID(field_value)
        # TODO(later): should this be more dynamic?
        elif field_name == 'message':
            return Message(**field_value)
        elif field_name == 'recipient':
            return Recipient(**field_value)
        else:
            return field_value


class MessageEncoder(json.JSONEncoder):
    """ Custom Message Encoder. """
    # The Pylint error is disabled because of a bug in Pylint.
    # See https://github.com/PyCQA/pylint/issues/414
    def default(self, o):
        if isinstance(o, UUID):
            return str(o)
        elif isinstance(o, date.datetime):
            return date.serialize(o)
        elif hasattr(o, 'to_json'):
            return o.to_json()
        elif isinstance(o, Promise):
            return str(o)
        else:
            return super().default(o)   # pragma: no cover



"""
:mod:`edx_ace.channel` exposes the ACE extension point needed
to add new delivery :class:`Channel` instances to an ACE application.

Developers wanting to add a new deliver channel should subclass :class:`Channel`,
and then add an entry to the ``openedx.ace.channel`` entrypoint in their ``setup.py``.
"""

import abc
import itertools
from collections import OrderedDict, defaultdict
from enum import Enum

from django.conf import settings

from edx_ace.errors import UnsupportedChannelError
from edx_ace.utils.once import once
from edx_ace.utils.plugins import get_plugins

# TODO(later): encapsulate the shared part of this namespace in the utils.plugin module
CHANNEL_EXTENSION_NAMESPACE = 'openedx.ace.channel'


class ChannelType(Enum):
    """
    All supported communication channels.
    """

    EMAIL = 'email'
    PUSH = 'push'

    def __str__(self):
        return str(self.value)


class Channel(metaclass=abc.ABCMeta):
    """
    Channels deliver messages to users that have already passed through the presentation and policy steps.

    Examples include email messages, push notifications, or in-browser messages. Implementations of this abstract class
    should not require any parameters be passed into their constructor since they are instantiated.

    :attr:`.channel_type` must be a :class:`.ChannelType`.
    """

    channel_type = None

    @classmethod
    def enabled(cls):
        """
        Validate settings to determine whether this channel can be enabled.
        """
        return True

    @abc.abstractmethod
    def deliver(self, message, rendered_message):
        """
        Transmit a rendered message to a recipient.

        Args:
            message (Message): The message to transmit.
            rendered_message (dict): The rendered content of the message that has been personalized for this particular
                recipient.
        """
        raise NotImplementedError()

    def overrides_delivery_for_message(self, message):  # pylint: disable=unused-argument
        """
        Returns true if this channel specifically wants to handle this message, outside normal channel delivery rules.

        For example, say you use a django transactional email channel, but with a default channel of braze.
        Then if the braze channel is configured with a campaign for a certain transactional message id specifically, it
        will claim that message via this method and end up delivering it via braze instead of the normal transactional
        django channel.
        """
        return False


class ChannelMap:
    """
    A class that represents a channel map, usually as described in Django settings and `setup.py` files.
    """

    def __init__(self, channels_list):
        """
        Initialize a ChannelMap.

        Args:
            channels_list (list): A list of [channel_name, channel] pairs to fill in the channel map, in order.
        """
        self.channel_type_to_channel_impl = defaultdict(OrderedDict)
        for channel_name, channel in channels_list:
            self.register_channel(channel, channel_name)

    def register_channel(self, channel, channel_name):
        """
        Registers a channel in the channel map.

        Args:
            channel (Channel): The channel to register.
            channel_name (str): The channel name, as stated in the `setup.py` file.
        """
        self.channel_type_to_channel_impl[channel.channel_type][channel_name] = channel

    def get_channel_by_name(self, channel_type, channel_name):
        """
        Gets a registered a channel by its name and type.

        Raises:
            KeyError: If either of the channel or its type are not registered.

        Returns:
            Channel: The channel object.
        """
        return self.channel_type_to_channel_impl[channel_type][channel_name]

    def get_default_channel(self, channel_type):
        """
        Returns the first registered channel by type.

        Raises:
            UnsupportedChannelError: If there's no channel that matched the request.

        Args:
            channel_type (ChannelType): The channel type.
        """
        try:
            return next(itertools.islice(self.channel_type_to_channel_impl[channel_type].values(), 0, 1))
        except (StopIteration, KeyError) as error:
            raise UnsupportedChannelError(
                f'No implementation for channel {channel_type} is registered. '
                f'Available channels are: {channels()}',
            ) from error

    def __str__(self):
        return f'ChannelMap {self.channel_type_to_channel_impl!r}'


@once
def channels():  # pragma: no cover
    """
    Gathers all available channels.

    Note that this function loads all available channels from entry points. It expects the Django setting
    ``ACE_ENABLED_CHANNELS`` to be a list of plugin names that should be enabled. Only one plugin per channel type
    should appear in that list.

    Raises:
        ValueError: If multiple plugins are enabled for the same channel type.

    Returns:
        ChannelMap: A mapping of channel types to instances of channel objects that can be used to deliver messages.
    """
    plugins = get_plugins(
        namespace=CHANNEL_EXTENSION_NAMESPACE,
        names=getattr(settings, 'ACE_ENABLED_CHANNELS', []),
    )

    return ChannelMap([
        [extension.name, extension.obj]
        for extension in plugins
    ])


def get_channel_for_message(channel_type, message):
    """
    Based on available `channels()` returns a single channels for a message.

    Raises:
        UnsupportedChannelError: If there's no channel matches the request.

    Returns:
        Channel: The selected channel object.
    """
    channels_map = channels()
    channel_names = []

    if channel_type == ChannelType.EMAIL:
        if message.options.get('override_default_channel'):
            channel_names = [message.options.get('override_default_channel')]
        elif message.options.get('transactional'):
            channel_names = [settings.ACE_CHANNEL_TRANSACTIONAL_EMAIL, settings.ACE_CHANNEL_DEFAULT_EMAIL]
        else:
            channel_names = [settings.ACE_CHANNEL_DEFAULT_EMAIL]
    elif channel_type == ChannelType.PUSH and getattr(settings, "ACE_CHANNEL_DEFAULT_PUSH", None):
        channel_names = [settings.ACE_CHANNEL_DEFAULT_PUSH]

    try:
        possible_channels = [
            channels_map.get_channel_by_name(channel_type, channel_name)
            for channel_name in channel_names
        ]
    except KeyError:
        return channels_map.get_default_channel(channel_type)

    # First see if any channel specifically demands to deliver this message
    for channel in possible_channels:
        if channel.overrides_delivery_for_message(message):
            return channel

    # Else the normal path: use the preferred channel for this message type
    if possible_channels:
        return possible_channels[0]
    else:
        return channels_map.get_default_channel(channel_type)



'''
edx-ace template tags
'''
from django import template

register = template.Library()


@register.simple_tag(takes_context=True)
def get_action_links(context, channel, omit_unsubscribe_link=False):
    """
    Retrieve the action links for the channel and pass the omit_unsubscribe_link argument
    """
    if getattr(channel, 'get_action_links', None):
        return channel.get_action_links(omit_unsubscribe_link=context.get('omit_unsubscribe_link'))
    return []



"""
:mod:`edx_ace.recipient_resolver` contains the :class:`RecipientResolver`, which facilitates
a design pattern that separates message content from recipient lists.
"""
from abc import ABCMeta, abstractmethod


# TODO(now): Do we even need to define this class? It seems like something the client could manage on their own.
class RecipientResolver(metaclass=ABCMeta):
    """
    This class represents a pattern for separating the content of a message
    (the :class:`.MessageType`) from the selection of recipients (the :class:`RecipientResolver`).
    """
    @abstractmethod
    def send(self, msg_type, *args, **kwargs):
        """
        :func:`.send` a :class:`.Message` personalized from ``msg_type`` to all
        recipients selected by this :class:`.RecipientResolver`.

        Arguments:
            msg_type (:class:`.MessageType`): An instantiated :class:`.MessageType`
                that describes the message batch to send.
        """



"""
:mod:`edx_ace.channel.mixins` implements some helper methods for channels
"""
import re

from django.conf import settings

from edx_ace.channel import ChannelType
from edx_ace.errors import FatalChannelDeliveryError


class EmailChannelMixin:
    """
    Adds some common email utility methods to email channels
    """
    channel_type = ChannelType.EMAIL

    @staticmethod
    def get_from_address(message):
        """Grabs the from_address from the message with fallback and error handling"""
        default_from_address = getattr(settings, 'DEFAULT_FROM_EMAIL', None)
        from_address = message.options.get('from_address', default_from_address)
        if not from_address:
            raise FatalChannelDeliveryError(
                'from_address must be included in message delivery options or as the DEFAULT_FROM_EMAIL settings'
            )
        return from_address

    @staticmethod
    def get_subject(rendered_message):
        # Compress spaces and remove newlines to make it easier to author templates.
        return re.sub('\\s+', ' ', rendered_message.subject, re.UNICODE).strip()

    @staticmethod
    def make_simple_html_template(head_html, body_html):
        return f"""<!DOCTYPE html>
<html>
  <head>
    {head_html}
  </head>
  <body>
    {body_html}
  </body>
</html>"""



"""
Tests for the PushNotificationChannel class.
"""
from unittest import TestCase
from unittest.mock import MagicMock, patch

import pytest
from firebase_admin.messaging import APNSConfig
from push_notifications.models import GCMDevice

from django.contrib.auth import get_user_model
from django.test import override_settings

from edx_ace.channel.push_notification import PushNotificationChannel
from edx_ace.errors import FatalChannelDeliveryError
from edx_ace.message import Message
from edx_ace.recipient import Recipient
from edx_ace.renderers import RenderedPushNotification
from edx_ace.utils.date import get_current_time

User = get_user_model()


@pytest.mark.django_db
class TestPushNotificationChannel(TestCase):
    """
    Tests for the PushNotificationChannel class.
    """

    def setUp(self):
        super().setUp()
        self.user = User.objects.create(username='username', email='email@example.com')
        self.lms_user_id = self.user.id
        self.msg_kwargs = {
            'app_label': 'test_app_label',
            'name': 'test_message',
            'expiration_time': get_current_time(),
            'context': {
                'key1': 'value1',
                'key2': 'value2',
            },
            'recipient': Recipient(
                lms_user_id=self.lms_user_id,
            )
        }

    @override_settings(PUSH_NOTIFICATIONS_SETTINGS={'CONFIG': 'push_notifications.conf.AppConfig'})
    def test_enabled(self):
        """
        Test that the channel is enabled when the settings are configured.
        """
        assert PushNotificationChannel.enabled()

    @override_settings(PUSH_NOTIFICATIONS_SETTINGS=None)
    def test_disabled(self):
        """
        Test that the channel is disabled when the settings are not configured.
        """
        assert not PushNotificationChannel.enabled()

    @patch('edx_ace.channel.push_notification.LOG')
    @patch('edx_ace.channel.push_notification.PushNotificationChannel.send_message')
    def test_deliver_no_device_tokens(self, mock_send_message, mock_log):
        """
        Test that the deliver method logs a message when the recipient has no push tokens.
        """
        with patch.object(PushNotificationChannel, 'get_user_device_tokens', return_value=[]):
            message = Message(options={}, **self.msg_kwargs)
            rendered_message = RenderedPushNotification(title='Test', body='This is a test.')

            channel = PushNotificationChannel()
            channel.deliver(message, rendered_message)

            mock_log.info.assert_called_with(
                'Recipient with ID %s has no push token. Skipping push notification.',
                self.lms_user_id
            )
            mock_send_message.assert_not_called()

    @patch('edx_ace.channel.push_notification.PushNotificationChannel.send_message')
    def test_deliver_with_device_tokens(self, mock_send_message):
        """
        Test that the deliver method sends a push notification for each device token.
        """
        with patch.object(PushNotificationChannel, 'get_user_device_tokens', return_value=['token1', 'token2']):
            message = Message(options={}, **self.msg_kwargs)
            rendered_message = RenderedPushNotification(title='Test', body='This is a test.')

            channel = PushNotificationChannel()
            channel.deliver(message, rendered_message)

            assert mock_send_message.call_count == 2

    @override_settings(FCM_APP_NAME='test_app')
    @patch('edx_ace.channel.push_notification.send_message')
    @patch('edx_ace.channel.push_notification.dict_to_fcm_message')
    @patch('edx_ace.channel.push_notification.PushNotificationChannel.collect_apns_config')
    def test_send_message_success(self, mock_collect_apns_config, mock_dict_to_fcm_message, mock_send_message):
        """
        Test that the send_message method sends a push notification successfully.
        """
        mock_dict_to_fcm_message.return_value = MagicMock()
        mock_collect_apns_config.return_value = MagicMock()

        message = Message(options={}, **self.msg_kwargs)
        rendered_message = RenderedPushNotification(title='Test', body='This is a test.')

        channel = PushNotificationChannel()
        channel.send_message(message, 'token', rendered_message)

        mock_send_message.assert_called_once()

    @override_settings(FCM_APP_NAME='test_app')
    @patch('edx_ace.channel.push_notification.send_message', side_effect=FatalChannelDeliveryError('Error'))
    @patch('edx_ace.channel.push_notification.dict_to_fcm_message')
    @patch('edx_ace.channel.push_notification.PushNotificationChannel.collect_apns_config')
    @patch('edx_ace.channel.push_notification.LOG')
    def test_send_message_failure(
        self, mock_log, mock_collect_apns_config, mock_dict_to_fcm_message, mock_send_message
    ):
        """
        Test that the send_message method logs an exception when an error occurs while sending the message.
        """
        mock_dict_to_fcm_message.return_value = MagicMock()
        mock_collect_apns_config.return_value = MagicMock()

        message = Message(options={}, **self.msg_kwargs)
        rendered_message = RenderedPushNotification(title='Test', body='This is a test.')

        channel = PushNotificationChannel()

        with pytest.raises(FatalChannelDeliveryError):
            channel.send_message(message, 'token', rendered_message)

        mock_send_message.assert_called_with('token', mock_dict_to_fcm_message.return_value, 'test_app')
        mock_log.exception.assert_called_with('Failed to send push notification to %s', 'token')

    def test_collect_apns_config(self):
        """
        Test that the collect_apns_config method returns an APNSConfig object with the correct headers.
        """
        notification_data = {'title': 'Test Title', 'body': 'Test Body'}

        apns_config = PushNotificationChannel.collect_apns_config(notification_data)

        assert isinstance(apns_config, APNSConfig)
        assert apns_config.headers['apns-priority'] == '5'
        assert apns_config.headers['apns-push-type'] == 'alert'

    def test_compress_spaces(self):
        """
        Test that the compress_spaces method removes extra spaces and newlines from a string.
        """
        compressed = PushNotificationChannel.compress_spaces('This   is a \n\n test.')
        assert compressed == 'This is a test.'

    def test_get_user_device_tokens(self):
        """
        Test that the get_user_device_tokens method returns the device tokens for a user.
        """
        gcm_device = GCMDevice.objects.create(user=self.user, registration_id='token1')

        channel = PushNotificationChannel()
        tokens = channel.get_user_device_tokens(self.lms_user_id)
        assert tokens == [gcm_device.registration_id]

    def test_get_user_device_tokens_no_tokens(self):
        """
        Test that the get_user_device_tokens method returns an empty list when the user has no device tokens.
        """
        channel = PushNotificationChannel()
        tokens = channel.get_user_device_tokens(self.lms_user_id)
        assert tokens == []



"""
:mod:`edx_ace.channel.django_email` implements a Django `send_mail()` email
delivery channel for ACE.
"""
import logging
from smtplib import SMTPException

from django.core.mail import EmailMultiAlternatives

from edx_ace.channel import Channel
from edx_ace.channel.mixins import EmailChannelMixin
from edx_ace.errors import FatalChannelDeliveryError

LOG = logging.getLogger(__name__)


class DjangoEmailChannel(EmailChannelMixin, Channel):
    """
    A `send_mail()` channel for edX ACE.

    This is both useful for providing an alternative to Sailthru and to debug ACE mail by
    inspecting `django.core.mail.outbox`.



    Example:

        Sample settings::

            .. settings_start
            EMAIL_BACKEND = 'django.core.mail.backends.smtp.EmailBackend'
            EMAIL_HOST = 'localhost'
            DEFAULT_FROM_EMAIL = 'hello@example.org'

            ACE_CHANNEL_DEFAULT_EMAIL = 'sailthru_email'
            ACE_CHANNEL_TRANSACTIONAL_EMAIL = 'django_email'

            ACE_ENABLED_CHANNELS = [
                'sailthru_email',
                'django_email',
            ]
            .. settings_end
    """

    @classmethod
    def enabled(cls):
        """
        Returns: True always!
        """
        return True

    def deliver(self, message, rendered_message):
        subject = self.get_subject(rendered_message)
        from_address = self.get_from_address(message)
        reply_to = message.options.get('reply_to', None)

        rendered_template = self.make_simple_html_template(rendered_message.head_html, rendered_message.body_html)
        try:
            mail = EmailMultiAlternatives(
                subject=subject,
                body=rendered_message.body,
                from_email=from_address,
                to=[message.recipient.email_address],
                reply_to=reply_to,
                headers=message.headers,
            )

            mail.attach_alternative(rendered_template, 'text/html')
            mail.send()
        except SMTPException as e:
            LOG.exception(e)
            raise FatalChannelDeliveryError('An SMTP error occurred (and logged) from Django send_email()') from e
